
import jahuwaldt.io.ExtFilenameFilter;
import jahuwaldt.js.tables.FTReaderFactory;
import jahuwaldt.js.tables.FTableReader;
import jahuwaldt.swing.AboutJMenuItem;
import jahuwaldt.swing.AppUtilities;
import jahuwaldt.swing.MDIApplication;
import jahuwaldt.swing.QuitJMenuItem;
import java.awt.Frame;
import java.awt.Window;
import java.awt.desktop.OpenFilesEvent;
import java.awt.desktop.QuitEvent;
import java.awt.desktop.QuitHandler;
import java.awt.desktop.QuitResponse;
import java.awt.event.ActionEvent;
import java.io.File;
import java.util.List;
import java.util.ResourceBundle;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JMenu;
import javax.swing.JMenuBar;


/**
 * This class represents a multi-document interface GUI application.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: September 3, 2011
 * @version January 19, 2024
 */
public class AppGUI extends MDIApplication {

    //  The logger for this class.
    private static final Logger LOGGER = Logger.getLogger(AppGUI.class.getName());
    
    /**
     * Constructor for our application that displays the GUI.
     * 
     * @param resBundle The resource bundle for this application.
     * @throws java.lang.Exception if anything can not be set up correctly.
     */
    @SuppressWarnings("OverridableMethodCallInConstructor")
    public AppGUI(ResourceBundle resBundle) throws Exception {
            super(resBundle.getString("appName"));
            
        // Set the system look and feel to hide that hideous Java LAF.
        AppUtilities.setSystemLAF();

        //	MacOS X users like the menu bar to be at the top of the screen.
        //	This does nothing on non-MacOS platforms.
        System.setProperty("apple.laf.useScreenMenuBar", "true");

        //	Store the resource bundle.
        setResourceBundle(resBundle);

        //  Get the list of data readers that are available.
        FTableReader[] allReaders = FTReaderFactory.getAllReaders();

        // Set up the file name filter.
        ExtFilenameFilter fnFilter = new ExtFilenameFilter();
        if (allReaders != null) {
            for (FTableReader reader : allReaders) {
                fnFilter.addExtension(reader.getExtension());
            }

            //  Add any other extensions that might be encountered that we should be able to read.
            fnFilter.addExtension("FDHS");
            fnFilter.addExtension("dat");
            fnFilter.addExtension("inp");
        }
        setFilenameFilter(fnFilter);

        //  Create the menu bar.
        JMenuBar menuBar = createMenuBar();

        // Add the menu bar to this application.
        //  Cause the application to stay open when the last window is closed under MacOS.
        //  When the last window is closed, a menu bar will remain for the application with no
        //  window open.  This is standard MacOS behavior and will do nothing on other platforms.
        setFramelessJMenuBar(menuBar);
        
        // Add an open files handler for MacOS.
        MDIApplication.setOpenFileHandler(new java.awt.desktop.OpenFilesHandler() {
            @Override
            public void openFiles(OpenFilesEvent e) {
                //  Get the list of files to open.
                List<File> files = e.getFiles();
                try {
                    Window tmp = getTopWindow();
                    if (tmp instanceof MainWindow) {
                        MainWindow topWindow = (MainWindow) tmp;

                        //  The top window is the right type. Use it.
                        String windowName = topWindow.getName();
                        for (File file : files) {
                            if (topWindow.openFile(file) > 0) {
                                // Build up a window name from the files opened.
                                if (!windowName.equals(""))
                                    windowName += ",";
                                windowName += file.getName();
                            }
                        }

                    } else {
                        //  Create a new application window and read in the files.
                        MainWindow.newWindowFromDataFiles(files);
                    }

                } catch (Exception ex) {
                    LOGGER.log(Level.SEVERE, null, ex);
                    AppUtilities.showException(null, MDIApplication.getInstance().getResourceBundle().getString("unexpectedTitle"),
                            MDIApplication.getInstance().getResourceBundle().getString("unexpectedMsg"), ex);
                }
            }
        });
        
    }

    //-----------------------------------------------------------------------------------
    /**
     * Handle the user requesting a new document window. Creates a new instance of this
     * application's main document window.
     *
     * @param event The event that caused this method to be called. May be "null" if this
     * method is called by MDIApplication or one of it's subclasses.
     * @return A new document window.
     */
    @Override
    public Frame handleNew(ActionEvent event) {
        MainWindow window = null;
        try {

            window = new MainWindow();
            window.setVisible(true);
            addWindow(window);

        } catch (Exception ex) {
            LOGGER.log(Level.SEVERE, null, ex);
            AppUtilities.showException(null, getResourceBundle().getString("unexpectedTitle"),
                    getResourceBundle().getString("unexpectedMsg"), ex);
        }

        return window;
    }

    /**
     * Handles the user choosing "Open" from the File menu. Allows the user to choose an
     * image file to open.
     * 
     * @param event The event that caused this method to be called.
     */
    public void handleOpen(ActionEvent event) {
        //  Create a new document window.
        MainWindow window = (MainWindow) handleNew(null);

        //  Have the new window handle opening the files.
        window.handleOpen(null);
    }

    /**
     * Initializes and displays the menus associated with this application.
     */
    private JMenuBar createMenuBar() throws NoSuchMethodException {
        ResourceBundle resBundle = getResourceBundle();

        //  Create a menu bar.
        JMenuBar menuBar = new JMenuBar();

        // Set up the file menu.
        int row = 0;
        String[][] menuStrings = new String[6][3];
        menuStrings[row][0] = resBundle.getString("newItemText");
        menuStrings[row][1] = resBundle.getString("newItemKey");
        menuStrings[row][2] = "handleNew";
        ++row;
        menuStrings[row][0] = resBundle.getString("openItemText");
        menuStrings[row][1] = resBundle.getString("openItemKey");
        menuStrings[row][2] = "handleOpen";
        ++row;
        ++row;	//	Blank line
        menuStrings[row][0] = resBundle.getString("closeItemText");
        menuStrings[row][1] = resBundle.getString("closeItemKey");
        menuStrings[row][2] = "handleClose";
        ++row;
        menuStrings[row][0] = resBundle.getString("saveItemText");
        menuStrings[row][1] = resBundle.getString("saveItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("saveAsItemText");
        menuStrings[row][1] = null;
        menuStrings[row][2] = null;

        JMenu menu = AppUtilities.buildMenu(this, resBundle.getString("fileMenuText"), menuStrings);
        menuBar.add(menu);

        //  Add a Quit menu item (if it isn't automatically present).
        if (!QuitJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(this.getQuitJMenuItem());
        }
        
        //  Add a Quit handler to ask the user if they want to save unsaved changes before quitting.
        setQuitHandler(new QuitHandler() {
            @Override
            public void handleQuitRequestWith(QuitEvent e, QuitResponse response) {
                handleQuit(e, response);
            }
        });
        
        
        // Set up the edit menu.
        row = 0;
        menuStrings = new String[6][3];
        menuStrings[row][0] = resBundle.getString("undoItemText");
        menuStrings[row][1] = resBundle.getString("undoItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("redoItemText");
        menuStrings[row][1] = null;
        menuStrings[row][2] = null;
        ++row;
        ++row;	//	Blank line.
        menuStrings[row][0] = resBundle.getString("cutItemText");
        menuStrings[row][1] = resBundle.getString("cutItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("copyItemText");
        menuStrings[row][1] = resBundle.getString("copyItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("pasteItemText");
        menuStrings[row][1] = resBundle.getString("pasteItemKey");
        menuStrings[row][2] = null;

        menu = AppUtilities.buildMenu(this, resBundle.getString("editMenuText"), menuStrings);
        menuBar.add(menu);

        //	Register an about menu item handler.
        setAboutHandler(AppUtilities.standardAboutHandler(resBundle, LOGGER));
        
        //  Create an about menu item (if not automatically present).
        if (!AboutJMenuItem.isAutomaticallyPresent()) {
            //  Create Help menu.
            menu = new JMenu(resBundle.getString("helpMenuText"));
            menuBar.add(menu);
            
            //  Add the "About" item to the Help menu.
            menu.add(this.getAboutJMenuItem());
        }
        
        return menuBar;
    }
}
