/*
 *  MainWindow  -- The main application window for the TableReader program.
 *
 *  Copyright (C) 2000-2024, Joseph A. Huwaldt.
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/gpl.html
 */

import jahuwaldt.io.ExtFilenameFilter;
import jahuwaldt.js.tables.FTReaderFactory;
import jahuwaldt.js.tables.FTableList;
import jahuwaldt.js.tables.FTableReader;
import jahuwaldt.js.tables.Float64Table;
import jahuwaldt.swing.*;
import java.awt.*;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.io.*;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.ResourceBundle;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.*;
import javax.swing.border.EmptyBorder;

/**
 * An application that can read, view, and write out a number of different functional
 * table formats.
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt Date: February 19, 2000
 * @version January 15, 2024
 */
public class MainWindow extends JFrame {

    //  The logger for this class.
    private static final Logger LOGGER = Logger.getLogger(MainWindow.class.getName());
    
    //  A count of the number of windows that have been created.
    private static int _windowCount = 0;

    // The list of tables shown in the GUI.
    private JList tableList = null;

    // The data model for the tableList.
    private final DefaultListModel listModel = new DefaultListModel();

    /**
     * A reference to the file that was last read in or written out to.
     */
    private File fileRef;

    /**
     * Database of all the tables that have been read in.
     */
    private FTableList allTables = null;

    //-----------------------------------------------------------------------------------
    /**
     * Constructs a new TableReader application window.
     *
     * @throws java.lang.NoSuchMethodException if there is any problem creating
     * this window instance.
     */
    @SuppressWarnings("OverridableMethodCallInConstructor")
    public MainWindow() throws NoSuchMethodException {
        super(MessageFormat.format(MDIApplication.getInstance().getResourceBundle().getString("mainTitle"), (_windowCount > 0 ? String.valueOf(_windowCount) : "")));

        //  Have the window dispose of itself when it closes.
        this.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
        this.setResizable(true);

        //  Define the window icon in the taskbar.
        ImageIcon icon = AppUtilities.getResourceIcon(MDIApplication.getInstance().getResourceBundle().getString("applicationIconResName"), LOGGER);
        if (icon != null)
            this.setIconImage(icon.getImage());

        // Layout the display of this window.
        layoutDisplay();

        //  Create a menu bar for this window.
        JMenuBar menuBar = createMenuBar();
        this.setJMenuBar(menuBar);

		//	Register a quit listener so that the user can save unsaved changes before the application quits.
        MDIApplication.addQuitListener(new QuitListener() {
            @Override
            public boolean quit() {
                return handleQuit();
            }
        });
        
        //  Size the window.
        this.pack();

        //  Position the window so that each new window can be seen.
        AppUtilities.positionWindow(this, getWidth(), getHeight());

        //  Increment the window count.
        ++_windowCount;
    }

    //-----------------------------------------------------------------------------------
    /**
     * Sets the title for this frame to the specified string. Also notifies the
     * main application class so that the "Windows" menu gets updated too.
     *
     * @param title The title to be displayed in the frame's border. A null
     * value is treated as an empty string, "".
     */
    @Override
    public void setTitle(String title) {
        super.setTitle(title);
        MDIApplication.windowTitleChanged(this);
    }

    /**
     * Layout the contents of this main application window.
     */
    private void layoutDisplay() throws NoSuchMethodException {
		ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();

        // Layout the main application window's user interface.
        JLabel selectLabel = new JLabel(resBundle.getString("selectTableLabel"), JLabel.LEFT);
        selectLabel.setAlignmentX(JLabel.LEFT_ALIGNMENT);

        // Set up buttons.
        JButton removeButton = new JButton(resBundle.getString("removeBtnText"));
        removeButton.addActionListener(AppUtilities.getActionListenerForMethod(this, "handleRemove"));
        removeButton.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
        removeButton.setToolTipText(resBundle.getString("removeBtnToolTip"));

        JButton addButton = new JButton(resBundle.getString("addBtnText"));
        addButton.addActionListener(AppUtilities.getActionListenerForMethod(this, "handleAdd"));
        addButton.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
        addButton.setToolTipText(resBundle.getString("addBtnToolTip"));

        JPanel addRemoveBtnPanel = new JPanel(new GridLayout(2, 1, 15, 15));
        addRemoveBtnPanel.setBorder(new EmptyBorder(20, 10, 20, 10));
        addRemoveBtnPanel.add(addButton);
        addRemoveBtnPanel.add(removeButton);

        JButton plotButton = new JButton(resBundle.getString("plotBtnText"));
        plotButton.addActionListener(AppUtilities.getActionListenerForMethod(this, "handlePlot"));
        plotButton.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
        plotButton.setToolTipText(resBundle.getString("plotBtnToolTip"));

        JButton lookupButton = new JButton(resBundle.getString("lookupBtnText"));
        lookupButton.addActionListener(AppUtilities.getActionListenerForMethod(this, "handleLookup"));
        lookupButton.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
        lookupButton.setToolTipText(resBundle.getString("lookupBtnToolTip"));

        JButton viewButton = new JButton(resBundle.getString("viewBtnText"));
        viewButton.addActionListener(AppUtilities.getActionListenerForMethod(this, "handleView"));
        viewButton.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
        viewButton.setToolTipText(resBundle.getString("viewBtnToolTip"));

        JPanel plotBtnPanel = new JPanel(new FlowLayout(FlowLayout.CENTER, 10, 5));
        plotBtnPanel.add(plotButton);
        plotBtnPanel.add(lookupButton);
        plotBtnPanel.add(viewButton);

        // Set up the table list.
        Box listBox = Box.createVerticalBox();
        listBox.setBorder(BorderFactory.createEmptyBorder(5, 20, 0, 0));
        listBox.add(selectLabel);

        tableList = new JList(listModel);
        tableList.setPrototypeCellValue("A Typical Run Name Entry Size");
        tableList.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
        JScrollPane scrollPane = new JScrollPane(tableList);
        scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        scrollPane.setAlignmentX(JScrollPane.LEFT_ALIGNMENT);
        listBox.add(scrollPane);

        // Set up application information.
        Box appInfoPanel = Box.createHorizontalBox();
        appInfoPanel.setBorder(new EmptyBorder(5, 20, 0, 10));
        JLabel appName = new JLabel(resBundle.getString("appName"));
        Font font = appName.getFont().deriveFont(Font.BOLD);
        font = font.deriveFont(font.getSize2D() * 1.2F);
        appName.setFont(font);
        JLabel version = new JLabel(
                MessageFormat.format(resBundle.getString("versionLabel"),
                        resBundle.getString("appVersion")));
        JLabel byLabel = new JLabel(
                MessageFormat.format(resBundle.getString("authoredByLabel"),
                        resBundle.getString("appAuthor")));
        Box vbox = Box.createVerticalBox();
        vbox.add(appName);
        vbox.add(version);
        vbox.add(byLabel);
        appInfoPanel.add(vbox);

        // Layout main content pane.
        Container cp = this.getContentPane();
        cp.add(appInfoPanel, BorderLayout.NORTH);
        cp.add(addRemoveBtnPanel, BorderLayout.EAST);
        cp.add(plotBtnPanel, BorderLayout.SOUTH);
        cp.add(listBox, BorderLayout.CENTER);

        //  Add support for drag and drop to the run list.
        new DropTarget(tableList, DnDConstants.ACTION_COPY, new FileDTListener(), true);
    }

    /**
     * A class that listens for Drag & Drop events that contain files. When a
     * drop event is received, this class reads in any files that were dropped.
     */
    private class FileDTListener extends DropTargetAdapter {

        /**
         * Called when a drag operation is ongoing, while the mouse pointer is
         * still over the operable part of the drop site for the
         * <code>DropTarget</code> registered with this listener. This
         * implementation will accept only COPY drop actions for
         * javaFileListFlavor flavors.
         */
        @Override
        public void dragOver(DropTargetDragEvent ev) {
            dropTargetDrag(ev);
        }

        /**
         * Called if the user has modified the current drop gesture. This
         * implementation will accept only COPY drop actions for
         * javaFileListFlavor flavors.
         */
        @Override
        public void dropActionChanged(DropTargetDragEvent ev) {
            dropTargetDrag(ev);
        }

        private void dropTargetDrag(DropTargetDragEvent ev) {
            //	Accept only file list compatible data.
            if (ev.isDataFlavorSupported(DataFlavor.javaFileListFlavor))
                ev.acceptDrag(DnDConstants.ACTION_COPY);    // Accept only a COPY type of transfer.
            else
                ev.rejectDrag();
        }

        /**
         * Called when the drag operation has terminated with a drop on the
         * operable part of the drop site for the <code>DropTarget</code>
         * registered with this listener. This implementation will process the
         * drop event, extract the file list and load each file in the list.
         */
        @Override
        public void drop(DropTargetDropEvent ev) {

            //	Accept only COPY drops.
            ev.acceptDrop(DnDConstants.ACTION_COPY);

            try {
                ev.getSource();
                List<File> files = (List<File>)ev.getTransferable().getTransferData(DataFlavor.javaFileListFlavor);

                //	Process the files, one at a time.
                int oldSize = 0;
                if (allTables != null)
                    oldSize = allTables.size();
                int addedCount = 0;
                for (File theFile : files) {

                    //	Is the file a real file (and not a directory)?
                    if (theFile.isFile()) {
                        //	Attempt to read in the data file.
                        FTableList newDB = readData(MainWindow.this, theFile, false, MDIApplication.getInstance().getResourceBundle());

                        //	Add the newly read in tables to this window's display.
                        addNewTables(newDB);
                        addedCount = newDB.size();
                    }
                }

                //	Select the runs that were added.
                if (addedCount > 0)
                    tableList.setSelectionInterval(oldSize, oldSize + addedCount - 1);

            } catch (UnsupportedFlavorException | IOException ex) {
                LOGGER.log(Level.SEVERE, null, ex);
            }

            ev.dropComplete(true);
        }
    }

    /**
     * Initializes the menus associated with this window.
     */
    private JMenuBar createMenuBar() throws NoSuchMethodException {
        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();

        JMenuBar menuBar = new JMenuBar();

        // Set up the file menu.
        int row = 0;
        String[][] menuStrings = new String[10][3];
        menuStrings[row][0] = resBundle.getString("newItemText");
        menuStrings[row][1] = resBundle.getString("newItemKey");
        menuStrings[row][2] = "handleNew";
        ++row;
        menuStrings[row][0] = resBundle.getString("openItemText");
        menuStrings[row][1] = resBundle.getString("openItemKey");
        menuStrings[row][2] = "handleOpen";
        ++row;
        ++row;	//	Blank line
        menuStrings[row][0] = resBundle.getString("closeItemText");
        menuStrings[row][1] = resBundle.getString("closeItemKey");
        menuStrings[row][2] = "handleClose";
        ++row;
        menuStrings[row][0] = resBundle.getString("saveAsItemText");
        menuStrings[row][1] = null;
        menuStrings[row][2] = null;
        int saveAsItem = row;			//	Store the item index of the Save As menu for later use.
        ++row;
        ++row;	//	Blank line
        menuStrings[row][0] = resBundle.getString("pageSetupItemText");
        menuStrings[row][1] = null;
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("printItemText");
        menuStrings[row][1] = resBundle.getString("printItemKey");
        menuStrings[row][2] = null;
        JMenu menu = AppUtilities.buildMenu(this, resBundle.getString("fileMenuText"), menuStrings);
        menuBar.add(menu);
        
        //	Add the Export As sub-menu.
        JMenu exportMenu = buildSaveAsMenu(menu.getItem(saveAsItem).getText());
        if (exportMenu != null) {
            menu.remove(saveAsItem);
            menu.add(exportMenu, saveAsItem);
        }
        
        //  Add a Quit menu item (if it isn't automatically present).
        AppGUI guiApp = (AppGUI)MDIApplication.getInstance();
        if (!QuitJMenuItem.isAutomaticallyPresent()) {
            menu.addSeparator();
            menu.add(guiApp.getQuitJMenuItem());
        }
        
        
        // Set up the edit menu.
        row = 0;
        menuStrings = new String[6][3];
        menuStrings[row][0] = resBundle.getString("undoItemText");
        menuStrings[row][1] = resBundle.getString("undoItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("redoItemText");
        menuStrings[row][1] = null;
        menuStrings[row][2] = null;
        ++row;
        ++row;	//	Blank line.
        menuStrings[row][0] = resBundle.getString("cutItemText");
        menuStrings[row][1] = resBundle.getString("cutItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("copyItemText");
        menuStrings[row][1] = resBundle.getString("copyItemKey");
        menuStrings[row][2] = null;
        ++row;
        menuStrings[row][0] = resBundle.getString("pasteItemText");
        menuStrings[row][1] = resBundle.getString("pasteItemKey");
        menuStrings[row][2] = null;
        menu = AppUtilities.buildMenu(this, resBundle.getString("editMenuText"), menuStrings);

        menuBar.add(menu);

		//  Create a Window's menu.
		menu = MDIApplication.newWindowsMenu(resBundle.getString("windowsMenuText"));
		menuBar.add(menu);
		
        //  Create an about menu item (if not automatically present).
        if (!AboutJMenuItem.isAutomaticallyPresent()) {
            //  Create Help menu.
            menu = new JMenu(resBundle.getString("helpMenuText"));
            menuBar.add(menu);
            
            //  Add the "About" item to the Help menu.
            menu.add(guiApp.getAboutJMenuItem());
        }
        
        return menuBar;
    }

    /**
     * Method that will build an "Export" menu containing a list of all the
     * FTableReader objects that are capable of writing to files.
     *
     * @param title	The title for the menu that is created.
     * @returns A JMenu instance containing a list of all FTableReader objects
     * that can write to files in various formats. If there are no appropriate
     * readers available, null is returned.
     */
    private JMenu buildSaveAsMenu(String title) {

        //  Get the list of data readers that are available.
        FTableReader[] allReaders = FTReaderFactory.getAllReaders();
        if (allReaders == null || allReaders.length < 1)
            return null;

        //	Create the menu
        JMenu menu = new JMenu(title);

        //  Loop over all the readers.
        boolean noneFound = true;
        for (FTableReader reader : allReaders) {

            if (reader.canWriteData()) {
                //  Found a "writer".
                noneFound = false;

                //  Create a menu item for it.
                JMenuItem menuItem = new JMenuItem(reader.toString() + "...");
                menuItem.addActionListener(new SaveAsMenuListener(reader));
                menuItem.setActionCommand(reader.toString());
                menuItem.setEnabled(true);

                //  Add the menu item to the menu.
                menu.add(menuItem);
            }
        }

        if (noneFound)
            menu = null;

        return menu;
    }

    /**
     * Defines an action listener for our save as menu items.
     */
    private class SaveAsMenuListener implements ActionListener {

        //  Keep a reference to the reader used by this menu.

        private final FTableReader reader;

        public SaveAsMenuListener(FTableReader reader) {
            this.reader = reader;
        }

        @Override
        public void actionPerformed(ActionEvent evt) {
            if (allTables != null && !allTables.isEmpty())
                handleSaveAs();
        }

        /**
         * Handles saving the data using this menu item's FTableReader instance.
         */
        private void handleSaveAs() {

            // Get a list of the items the user has selected.
            int[] selectionIndices = tableList.getSelectedIndices();
            if (selectionIndices == null || selectionIndices.length < 1)
                return;

            ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();

            //	Set up the file name filter.
            String extension = reader.getExtension();
            ExtFilenameFilter filter = new ExtFilenameFilter(extension, reader.toString() + " File");

            //  Build up a file name.
            String dir = MDIApplication.getInstance().getPreferences().getLastPath();
            String fileName = resBundle.getString("untitled");
            if (fileRef != null)
                fileName = fileRef.getName();
            int idx = fileName.lastIndexOf(".");
            if (idx > 0)
                fileName = fileName.substring(0, idx);
            fileName += "." + extension;

            //	Ask the user to select a file for saving.
            File chosenFile = AppUtilities.selectFile4Save(MainWindow.this,
                    MessageFormat.format(resBundle.getString("fileSaveDialogMsg"), reader.toString()),
                    dir, fileName, filter, extension,
                    resBundle.getString("fileExistsMsgFmt"), resBundle.getString("warningTitle"));

            if (chosenFile != null) {
                if (chosenFile.exists() && !chosenFile.canWrite()) {
                    JOptionPane.showMessageDialog(MainWindow.this,
                            resBundle.getString("canNotWrite2File"),
                            resBundle.getString("ioErrorTitle"), JOptionPane.ERROR_MESSAGE);
                    return;
                }

                //	Export the data.
                exportData(selectionIndices, reader, chosenFile);

                //	Store off the path reference for later use.
                fileRef = chosenFile;
                MDIApplication.getInstance().getPreferences().setLastPath(chosenFile.getParent());
            }

        }

        /**
         * Method that exports the specified data to the specified file using
         * the specified writer.
         *
         * @param selection A list of the indices of the selected tables to be
         * saved.
         * @param writer The FTableReader to use to write out the data.
         * @param theFile Reference to the file path where the file should be
         * written.
         */
        private void exportData(int[] selection, FTableReader writer, File theFile) {

            // Create a new table database containing only the selected tables.
            FTableList theTables = FTableList.newInstance();
            for (int item : selection) {
                theTables.add((Float64Table) allTables.get(item));
            }

            // Copy over the notes.
            theTables.addAllNotes(allTables.getAllNotes());

            try (OutputStream output = new FileOutputStream(theFile)) {
                // Write out the selected tables using the specified writer.
                writer.write(output, theTables);

            } catch (IOException | ArrayIndexOutOfBoundsException e) {
                theFile.delete();
                LOGGER.log(Level.SEVERE, null, e);
                ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
                JOptionPane.showMessageDialog(MainWindow.this, e.getMessage(), resBundle.getString("ioErrorTitle"),
                        JOptionPane.ERROR_MESSAGE);
            }
        }

    }   //  end class SaveAsMenuListener

    /**
     * Function that opens the specified data file and adds any tables in it to those in
     * the current window.
     *
     * @param theFile The data file to read in.
     * @return A count of how many tables were read in from the file.
     */
    public int openFile(File theFile) {

        // Read the file and add it to the current window.
        FTableList data = readData(this, theFile, false, MDIApplication.getInstance().getResourceBundle());

        int addedCount = 0;
        if (data != null) {
            // Add the new tables to the new window.
            addNewTables(data);
            addedCount += data.size();
        }
        
        return addedCount;
    }

    /**
     * Method that reads in the specified file and creates a new window for displaying
     * it's contents.
     *
     * @param theFiles The list of files to be loaded and displayed. If null is passed,
     *                 this method will simply create an empty window.
     * @return A reference to the new window that is created and displayed.
     * @throws java.lang.NoSuchMethodException if there is a problem creating the window
     * instance.
     */
    public static MainWindow newWindowFromDataFiles(List<File> theFiles)
            throws NoSuchMethodException {
        if (theFiles == null)
            theFiles = new ArrayList();

        // Create an instance of this window.
        MainWindow window = new MainWindow();

        String windowName = "";

        // Add the specified data files to the new window.
        for (File file : theFiles) {
            if (window.openFile(file) > 0) {
                // Build up a window name from the files opened.
                if (!windowName.equals(""))
                    windowName += ",";
                windowName += file.getName();
            }
        }

        if (!windowName.equals(""))
            window.setName(windowName);
        window.setVisible(true);
        MDIApplication.addWindow(window);

        return window;
    }

    /**
     * Add a new table database to the list of tables in this window.
     *
     * @param newDB A new list of data tables.
     */
    public void addNewTables(FTableList newDB) {
        if (newDB != null) {
            if (allTables == null) {
                // Window currently has no data in it.
                allTables = newDB;

                // Add all the new tables to our displayed list.
                listModel.clear();
                for (Float64Table newTable : allTables) {
                    String label = createListEntrtyLabel(newTable);
                    listModel.addElement(label);
                }

            } else {
                // Add to existing window's data.
                for (Float64Table newTable : newDB) {
                    allTables.add(newTable);
                    String label = createListEntrtyLabel(newTable);
                    listModel.addElement(label);
                }
            }

            // Update the display.
            this.repaint();
        }
    }

    /**
     * Create a label for the list entry for a specific table. The label
     * contains the name of the table and the number of breakpoints in each
     * dimension.
     */
    private static String createListEntrtyLabel(Float64Table table) {

        StringBuilder label = new StringBuilder(table.getName().toString());
        label.append("  - (");
        int dims = table.getDimensions();
        for (int dim = 0; dim < dims; ++dim) {
            if (dim > 0)
                label.append("x");
            label.append(table.getNumBreakpoints(dim));
        }
        label.append(")");

        return label.toString();
    }

    /**
     * Method that reads in a data file and return the resulting data structure.
     *
     * @param parent The parent frame for dialogs (null is fine).
     * @param theFile The file to be read in. If null is passed, this method
     * will do nothing.
     * @param skipUnknown If <code>true</code>, then files with unknown types
     * will be silently ignored. If <code>false</code>, then files with unknown
     * types are reported to the user.
     * @param resBundle The resource bundle for this application.
     */
    private static FTableList readData(Frame parent, File theFile, boolean skipUnkown, ResourceBundle resBundle) {
        if (theFile == null)
            return null;

        FTableList tables = null;
        InputStream is = null;
        try {
            if (!theFile.exists())
                throw new IOException(
                        MessageFormat.format(resBundle.getString("fileDoesntExistErr"),
                                theFile.getName()));

            //  Get a reader for the selected file.
            FTableReader reader = FTReaderFactory.getReader(theFile);
            if (reader == null) {
                if (skipUnkown)
                    return null;
                throw new IOException(
                        MessageFormat.format(resBundle.getString("readerNotFound"),
                                theFile.getName()));
            }

            // The user has selected one of the formats we support.
            is = new FileInputStream(theFile);
            tables = reader.read(is);

        } catch (FileNotFoundException e) {
            JOptionPane.showMessageDialog(parent,
                    MessageFormat.format(resBundle.getString("fileNotFoundMsg"), theFile.getName()),
                    resBundle.getString("fileNotFoundTitle"), JOptionPane.ERROR_MESSAGE);
        } catch (IOException e) {
            LOGGER.log(Level.SEVERE, null, e);
            JOptionPane.showMessageDialog(parent, e.getMessage(), resBundle.getString("ioErrorTitle"),
                    JOptionPane.ERROR_MESSAGE);
        } finally {
            //  Make sure we close the stream no matter what happens.
            if (is != null) {
                try {
                    is.close();
                } catch (IOException ignore) { }
            }
        }

        return tables;
    }

	/**
	*  Method called by the quit listener when the application has been asked to quit.
	*  Returns true if the quit should be canceled, or false if continuing with
	*  quitting the application is fine.
	*/
	private boolean handleQuit() {
		return false;
	}
	
    /**
     * Handles the user choosing "New..." from the File menu. Creates a new
     * application window instance.
     * 
     * @param event The action that caused this method to be called.  Ignored.
     * @return A new MainWindow.
     */
    public MainWindow handleNew(ActionEvent event) {
        return (MainWindow)(MDIApplication.getInstance().handleNew(event));
    }

    /**
     * Handle the user choosing "Close" from the File menu. This implementation
     * dispatches a "Window Closing" event to this window.
     * 
     * @param event The action that caused this method to be called.  Ignored.
     */
    public void handleClose(ActionEvent event) {
        this.dispatchEvent(new WindowEvent(this, WindowEvent.WINDOW_CLOSING));
    }

    /**
     * Handle the user clicking on the "Add" button. Reads in a table file
     * selected by the user and adds those tables to the ones already in memory
     * (if any).
     * 
     * @param event The action that caused this method to be called.
     */
    public void handleAdd(ActionEvent event) {
        handleOpen(event);
    }

    /**
     * Handle the user clicking on the "Remove" button. Removes the selected
     * tables from the list of currently open tables.
     * 
     * @param event The action that caused this method to be called.  Ignored.
     */
    public void handleRemove(ActionEvent event) {

        // If only 1 table is loaded, select it automatically.
        if (listModel.size() == 1)
            tableList.setSelectedIndex(0);

        if (!tableList.isSelectionEmpty()) {

            // Get a list of the items the user has selected.
            int[] selectionIndices = tableList.getSelectedIndices();
            int length = selectionIndices.length;

            // Create a message to display to the user.
            ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
            StringBuilder buffer = new StringBuilder("<html>");
            buffer.append(resBundle.getString("removeConfirmText"));
            buffer.append("<br>");
            for (int i = 0; i < length; ++i) {
                Float64Table tbl = allTables.get(selectionIndices[i]);
                String item = tbl.getName().toString();
                buffer.append(item);
                if (i < length - 1) {
                    buffer.append(", ");
                    if (i > 1 && i % 5 == 0)
                        buffer.append("<br/>");
                }
            }
            buffer.append("</html>");

            // Ask the user for confirmation.
            int result = JOptionPane.showConfirmDialog(this, buffer.toString(), resBundle.getString("warningTitle"),
                    JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE);
            if (result == JOptionPane.YES_OPTION) {
                // Get the indicies for the selected rows of the table.
                int[] rows = tableList.getSelectedIndices();

                // The user has confirmed, so remove the selected tables from database and list.
                length = rows.length;
                for (int i = length - 1; i >= 0; --i) {
                    int row = rows[i];
                    allTables.remove(row);
                    listModel.remove(row);
                }

                if (allTables.isEmpty())
                    // No more tables, allow garbage collection on table database.
                    allTables = null;

                tableList.clearSelection();

                // Update the display.
                this.repaint();
            }
        }

    }

    /**
     * Handle the user clicking on the "Lookup" button. Opens a lookup window
     * and lets the user interactively lookup values in the table.
     * 
     * @param event The action that caused this method to be called.  Ignored.
     */
    public void handleLookup(ActionEvent event) {
        try {
            //	If only 1 table is loaded, select it automatically.
            if (listModel.size() == 1)
                tableList.setSelectedIndex(0);

            if (!tableList.isSelectionEmpty()) {

                // Get a list of the items the user has selected.
                int[] selection = tableList.getSelectedIndices();

                for (int item : selection) {
                    // Create an instance of a lookup window for each selected table.
                    Float64Table theTable = allTables.get(item);
                    LookupWindow lookup = new LookupWindow(theTable);
                    lookup.setVisible(true);
                    MDIApplication.addWindow(lookup);
                }
            }
        } catch (NoSuchMethodException e) {
            LOGGER.log(Level.SEVERE, null, e);
            ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
            AppUtilities.showException(null, resBundle.getString("noSuchMethodTile"),
                    resBundle.getString("unexpectedMsg"), e);
        }
    }

    /**
     * Handle the user clicking on the "Plot" button. Opens a plot window and
     * displays the currently selected table's data.
     * 
     * @param event The action that caused this method to be called.  Ignored.
     */
    public void handlePlot(ActionEvent event) {
        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
        try {
            //	If only 1 table is loaded, select it automatically.
            if (listModel.size() == 1)
                tableList.setSelectedIndex(0);

            if (!tableList.isSelectionEmpty()) {

                // Get a list of the items the user has selected.
                int[] selection = tableList.getSelectedIndices();

                for (int item : selection) {
                    // Create an instance of a plot window for each selected table.
                    Float64Table theTable = allTables.get(item);
                    String title = MessageFormat.format(resBundle.getString("plotText"), theTable.getName());
                    TablePlotWindow plot = new TablePlotWindow(title, theTable);
                    plot.setVisible(true);
                    MDIApplication.addWindow(plot);
                }
            }
        } catch (NoSuchMethodException e) {
            LOGGER.log(Level.SEVERE, null, e);
            AppUtilities.showException(null, resBundle.getString("noSuchMethodTile"),
                    resBundle.getString("unexpectedMsg"), e);
        }
    }

    /**
     * Handle the user clicking on the "View" button. Opens a view window and
     * displays the currently selected table's data.
     * 
     * @param event The action that caused this method to be called.  Ignored.
     */
    public void handleView(ActionEvent event) {
        try {
            //	If only 1 table is loaded, select it automatically.
            if (listModel.size() == 1)
                tableList.setSelectedIndex(0);

            if (!tableList.isSelectionEmpty()) {

                // Get a list of the items the user has selected.
                int[] selection = tableList.getSelectedIndices();

                for (int item : selection) {
                    // Create an instance of a view window for each selected table.
                    Float64Table theTable = allTables.get(item);
                    TableViewWindow window = new TableViewWindow(theTable);
                    window.setVisible(true);
                    MDIApplication.addWindow(window);
                }
            }
        } catch (NoSuchMethodException e) {
            LOGGER.log(Level.SEVERE, null, e);
            ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
            AppUtilities.showException(null, resBundle.getString("noSuchMethodTile"),
                    resBundle.getString("unexpectedMsg"), e);
        }
    }

    /**
     * Handle the user choosing "Open" from the File menu. Reads in a table file
     * selected by the user and replaces any existing tables with the ones from
     * that file.
     * 
     * @param event The action that caused this method to be called.  Ignored.
     */
    public void handleOpen(ActionEvent event) {
        String dir = MDIApplication.getInstance().getPreferences().getLastPath();
        if (dir != null && fileRef != null)
            dir = fileRef.getParent();
        openFileDialog(dir);
    }

    /**
     * Present the user with an open file dialog in the specified directory.
     * 
     * @param dir The directory path to prompt the user with in the open dialog.
     */
    public void openFileDialog(String dir) {
        ResourceBundle resBundle = MDIApplication.getInstance().getResourceBundle();
        try {
            File theFile = AppUtilities.selectFile(this, FileDialog.LOAD, resBundle.getString("fileDialogLoad"), dir,
                    "", MDIApplication.getFilenameFilter());
            if (theFile == null)
                return;

            //	Read the file and add it to the current window.
            FTableList newDB = readData(this, theFile, false, resBundle);
            fileRef = theFile;

            //	Store off the path reference for later use.
            MDIApplication.getInstance().getPreferences().setLastPath(theFile.getParent());

            if (newDB != null) {
                //	Add the new ones to the existing list of tables.
                addNewTables(newDB);
            }

        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, null, e);
            AppUtilities.showException(null, resBundle.getString("unexpectedTitle"),
                    resBundle.getString("unexpectedMsg"), e);
        }
    }

}
