/*
 *  TableReader  -- A Java Application for reading, viewing, and writing out functional tables.
 *
 *  Copyright (C) 2000-2024, Joseph A. Huwaldt
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/gpl.html
 */

import jahuwaldt.swing.AppUtilities;
import jahuwaldt.swing.MDIApplication;
import jahuwaldt.swing.MainApp;
import jahuwaldt.swing.Preferences;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;

/**
 * An application that can read, view, and write out a number of different functional
 * table formats.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt Date: February 19, 2000
 * @version January 15, 2024
 */
public class TableReader implements MainApp {

    /**
     * The resource bundle containing the string resources for this application. Resource
     * bundles are useful for localizing applications. New localities (languages) can be
     * added by adding properties files only -- no code modifications.
     */
    public static final ResourceBundle RESBUNDLE = ResourceBundle.getBundle("appStrings", Locale.getDefault());
    
    //	A reference to our GUI application instance.
    private final MDIApplication guiApp;
    
    /**
     * The user preferences for this program.
     */
    private final AppPreferences prefs;
    

    //-----------------------------------------------------------------------------------
    /**
     * Entry point for this application. This is where it starts up from.
     * 
     * @param args Command line arguments.
     */
    public static void main(final String[] args) {

        // Schedule a job for the event-dispatching thread:
        // creating and showing this application's GUI.
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                createAndShowGUI(args);
            }
        });

    }

    /**
     * Create the GUI and show it. For thread safety, this method should be invoked from
     * the event-dispatching thread.
     */
    private static void createAndShowGUI(String[] args) {

        try {
            // Create an instance of our application to get things rolling.
            new TableReader(args);

        } catch (NoSuchMethodException e) {
            e.printStackTrace();
            AppUtilities.showException(null, RESBUNDLE.getString("noSuchMethodTile"),
                    RESBUNDLE.getString("unexpectedMsg"), e);
            System.exit(0);

        } catch (Throwable e) {
            e.printStackTrace();
            AppUtilities.showException(null, RESBUNDLE.getString("unexpectedTitle"),
                    RESBUNDLE.getString("unexpectedMsg"), e);
            System.exit(0);
        }

    }

    /**
     * Construct a new TableReader application open and empty window.
     * 
     * @param args Command line arguments.
     * @throws java.lang.NoSuchMethodException if there is any problem starting
     * up the program.
     */
    @SuppressWarnings("LeakingThisInConstructor")
    public TableReader(String[] args) throws NoSuchMethodException, Exception {

        guiApp = new AppGUI(RESBUNDLE);

        //	Load in the user preferences.
        prefs = new AppPreferences(this);
        MDIApplication.setPreferences(prefs);

        if (args.length > 0) {
            //	Deal with the command line arguments.

            //	Return a list of arguments without the options.
            List<String> argsNoOptions = argumentsWithoutOptions(args);

            if (argumentStartsWith(args, "-h") >= 0 || argumentStartsWith(args, "--help") >= 0)
                printHelpAndDie();

            if (argumentStartsWith(args, "-v") >= 0) {
                System.out.print(RESBUNDLE.getString("appName") + " " + RESBUNDLE.getString("version") + " ");
                System.out.println(RESBUNDLE.getString("appVersion"));
                System.out.println();
            }

            //	Create an instance of the application window using the supplied data files.
            MainWindow window = null;
            if (!argsNoOptions.isEmpty()) {
                List<File> files = new ArrayList();
                for (String filename : argsNoOptions) {
                    File file = new File(filename);
                    files.add(file);
                }
                window = MainWindow.newWindowFromDataFiles(files);
            }

            int idx = argumentStartsWith(args, "--dir");
            if (idx >= 0) {
                //	The user wants to have the user select a file from the specified directory.

                //	Parse out the file path name.
                String path = args[idx + 1].trim();
                File dir = new File(path);
                if (dir.exists()) {

                    //	Create a new window if it hasn't been created already.
                    if (window == null) {
                        window = new MainWindow();
                        window.setVisible(true);
                        MDIApplication.addWindow(window);
                    }

                    //	Bring up the file selection dialog.
                    window.openFileDialog(dir.getAbsolutePath());
                }

            }

            //	If the command-line options did not include opening any files, then exit.
            if (idx < 0 && argsNoOptions.size() <= 0)
                System.exit(0);


        } else {
            //	No command line options, just create an empty window.
            //  Create a main window frame for this application.
            MainWindow window = new MainWindow();
            window.setVisible(true);
            MDIApplication.addWindow(window);
        }
    }

    /**
     * Method that prints a help message to standard out and then exits the program.
     */
    private static void printHelpAndDie() {
        System.out.println();
        System.out.println(RESBUNDLE.getString("cmdLineHelp"));
        System.out.println();
        System.exit(0);
    }

    /**
     * Returns the index of of the argument that starts with the specified string
     * (ignoring case) or -1 if the argument is not found.
     */
    private int argumentStartsWith(String[] args, String value) {
        value = value.toLowerCase();
        int length = args.length;
        for (int i = 0; i < length; ++i) {
            if (args[i].toLowerCase().startsWith(value))
                return i;
        }
        return -1;
    }

    /**
     * Takes the input list of arguments and strips out any arguments that are options
     * (i.e.: start with "-").
     */
    private List<String> argumentsWithoutOptions(String[] args) {
        List<String> list = new java.util.ArrayList();
        int length = args.length;
        for (int i = 0; i < length; ++i) {
            if (!args[i].startsWith("-"))
                list.add(args[i]);

            //	Skip the next argument if this one starts with "--".
            if (args[i].startsWith("--"))
                ++i;
        }
        return list;
    }

    /**
     * Return the index to the argument that starts with the specified value (ignoring
     * case).
     *
     * @param args The list of all the arguments.
     * @param value The value we are looking for.
     * @return the index to the argument that starts with the specified value ignoring
     *      case or -1 if none could be found.
     */
    private int argumentIndex(String[] args, String value) {
        value = value.toLowerCase();
        int length = args.length;
        for (int i = 0; i < length; ++i) {
            if (args[i].toLowerCase().startsWith(value))
                return i;
        }
        return -1;
    }

    /**
     * Returns the resource bundle stored with this application. If no resource bundle has
     * been stored, then null is returned.
     * 
     * @return The resource bundle for this program.
     */
    @Override
    public final ResourceBundle getResourceBundle() {
        return RESBUNDLE;
    }

    /**
     * Return a reference to this program's GUI application or null if we are running in
     * batch mode.
     * 
     * @return The GUI application instance for this program.
     */
    @Override
    public final MDIApplication getGUIApplication() {
        return guiApp;
    }

    /**
     * Return a reference to the user preferences for this application.
     * 
     * @return The preferences for this program.
     */
    @Override
    public final Preferences getPreferences() {
        return prefs;
    }
}
