/*
 *   C81Reader  -- A class that can read C-81 airfoil aerodynamic tables.
 *
 *   Copyright (C) 2003-2024, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.tables;

import jahuwaldt.io.FileUtils;
import java.io.*;
import java.text.DecimalFormat;
import java.text.MessageFormat;
import static java.text.MessageFormat.format;
import java.text.NumberFormat;
import java.text.ParseException;
import java.util.Locale;
import static java.util.Objects.requireNonNull;

/**
 * A class that provides a method for reading C-81 airfoil aerodynamics data tables.
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt, Date: November 13, 2002
 * @version January 15, 2024
 */
public class C81Reader implements FTableReader {

    //  Debugging flag.
    private static final boolean DEBUG = false;

    //  The preferred file extension for files of this reader's type.
    private static final String EXTENSION = "c81";
    
    /**
     * Returns a string representation of the object. This will return a brief description
     * of the format read by this reader.
     */
    @Override
    public String toString() {
        return RESOURCES.getString("c81Description");
    }

    /**
     * Returns the preferred file extension (not including the ".") for files of this
     * reader's type.
     * 
     * @return The preferred file extension ("c81").
     */
    @Override
    public String getExtension() {
        return EXTENSION;
    }

    /**
     * Method that determines if this reader can read data from the specified input
     * stream. Will return FTableReader.MAYBE if the file name has the extension "c81".
     *
     * @param name  The name of the file.
     * @param input An input stream containing the data to be read. Any methods that read
     *              from this stream must first set a mark and then reset back to that
     *              mark before the method returns (even if it returns with an exception).
     * @throws java.io.IOException
     * @returns FTableReader.NO if the file is not recognized at all or FTableReader.MAYBE
     * if the filename extension is "*.c81".
     */
    @Override
    public int canReadData(String name, BufferedInputStream input) throws IOException {
        requireNonNull(name, MessageFormat.format(RESOURCES.getString("paramNullErr"), "name"));

        int response = NO;
        String ext = FileUtils.getExtension(name);
        if (EXTENSION.equals(ext))
            response = MAYBE;

        return response;
    }

    /**
     * Returns false. This class can NOT write any data to a C-81 airfoil aerodynamics
     * formatted file at this time.
     * 
     * @return false
     */
    @Override
    public boolean canWriteData() {
        return false;
    }

    /**
     * Compares this object with the specified object for order based on the
     * <code>toString().compareTo(o.toString())</code> method. Returns a negative integer,
     * zero, or a positive integer as this object is less than, equal to, or greater than
     * the specified object.
     */
    @Override
    public int compareTo(FTableReader o) {
        return this.toString().compareTo(o.toString());
    }

    /**
     * Method that reads in C-81 airfoil aerodynamics formatted data from the specified
     * input stream and returns that data as an FTableList object.
     *
     * @param input An input stream containing the C-81 formatted data.
     * @returns An FTableList object that contains the data read in from the specified
     * stream.
     * @throws IOException If there is a problem reading the specified stream.
     */
    @Override
    public FTableList read(InputStream input) throws IOException {
        requireNonNull(input, MessageFormat.format(RESOURCES.getString("paramNullErr"), "input"));
 
        //  C-81 files are required to be in ASCII with U.S. style number formatting.
        //  Get the default locale and save it.  Then change the default locale to US.
        Locale defLocale = Locale.getDefault();
        Locale.setDefault(Locale.US);

        // Create a reader to translate the input stream.
        try {
            LineNumberReader reader = new LineNumberReader(new InputStreamReader(input));

            //  Parse the file's contents.
            FTableList db = readC81(reader);

            return db;
            
        } finally {
            //  Restore the default locale.
            Locale.setDefault(defLocale);
        }

    }

    /**
     * Method for writing out all the data stored in the specified list of FloatTable
     * objects to the specified output stream. This method is not implemented and will
     * throw an IOException if called.
     *
     * @param output The output stream to which the data is to be written.
     * @param tables The list of FloatTable objects to be written out.
     * @throws IOException If there is a problem writing to the specified stream.
     */
    @Override
    public void write(OutputStream output, FTableList tables) throws IOException {
        throw new IOException(format(RESOURCES.getString("writeNotImplementedMsg"), "C-81"));
    }

    /**
     * Read in a C-81 airfoil aero data table file using the specified reader.
     *
     * @param reader A file reader that reads in the characters from the file.
     * @return A collection of tables read in using the reader.
     */
    private static FTableList readC81(LineNumberReader reader) throws IOException {

        //  Start reading lines looking for the start of the data.
        String aLine = reader.readLine();
        while (aLine != null && aLine.length() < 34 && !aLine.regionMatches(0, " ", 0, 1)) {
            aLine = reader.readLine();
        }
        if (aLine == null)
            throw new IOException(RESOURCES.getString("c81MissingFormat"));

        // Create an empty table database.
        FTableList db = FTableList.newInstance();

        //  Parse out any notes.
        String buffer = aLine.substring(0, 29).trim();
        if (buffer.length() > 1)
            db.addNote(buffer);

        try {
            //  Create a number format for parsing our numbers.
            DecimalFormat nf = (DecimalFormat)NumberFormat.getInstance();

            String tableName;
            String[] indepNames = {"Mach", "Alpha"};

            int[] numMachs = new int[3];
            int[] numAlphas = new int[3];

            //  Extract the number of breakpoints in each table.
            int pos = 30;
            for (int tbl = 0; tbl < 3; ++tbl, pos += 2) {

                //  Extract the number of Mach breakpoints in table #0.
                buffer = aLine.substring(pos, pos + 2).trim();
                numMachs[tbl] = nf.parse(buffer).intValue();
                if (aLine == null)
                    throw new IOException(format(RESOURCES.getString("EOFErrMsg"), reader.getLineNumber()));

                //  Extract the number of alpha breakpoints in table #0.
                pos += 2;
                buffer = aLine.substring(pos, pos + 2).trim();
                numAlphas[tbl] = nf.parse(buffer).intValue();

            }

            //  Loop over the 3 tables to be read in (Cl, Cd, & Cm).
            for (int tbl = 0; tbl < 3; ++tbl) {

                //  Allocate arrays needed to build tables.
                double[][] depData = new double[numMachs[tbl]][numAlphas[tbl]];
                double[] machBP = new double[numMachs[tbl]];
                double[] alphaBP = new double[numAlphas[tbl]];

                //  Read in the Mach number breakpoints.
                aLine = reader.readLine();
                if (aLine == null)
                    throw new IOException(format(RESOURCES.getString("EOFErrMsg"), reader.getLineNumber()));

                if (DEBUG) {
                    System.out.println("Reading table #" + tbl);
                    System.out.println("numMachs = " + numMachs[tbl] + ", numAlphas = " + numAlphas[tbl]);
                    System.out.print("machBP[] =");
                }

                pos = 7;
                for (int i = 0; i < numMachs[tbl]; ++i, pos += 7) {
                    if (pos > 64) {
                        aLine = reader.readLine();
                        if (aLine == null)
                            throw new IOException(format(RESOURCES.getString("EOFErrMsg"), reader.getLineNumber()));
                        pos = 7;
                    }
                    buffer = aLine.substring(pos, pos + 7).trim();
                    double value = nf.parse(buffer).doubleValue();
                    machBP[i] = value;

                    if (DEBUG)
                        System.out.print(" " + value);
                }
                if (DEBUG) {
                    System.out.println();
                    System.out.print("alphaBP[] =");
                }

                //  Start reading in the alpha breakpoints and the dependent data.
                for (int i = 0; i < numAlphas[tbl]; ++i) {

                    //  Read in the next alpha line #1.
                    aLine = reader.readLine();
                    if (aLine == null)
                        throw new IOException(format(RESOURCES.getString("EOFErrMsg"), reader.getLineNumber()));

                    //  Parse out the alpha.
                    buffer = aLine.substring(0, 7).trim();
                    double value = nf.parse(buffer).doubleValue();
                    alphaBP[i] = value;

                    if (DEBUG)
                        System.out.print(" " + value);

                    //  Parse out the dependent data as a function of Mach number.
                    pos = 7;
                    for (int j = 0; j < numMachs[tbl]; ++j, pos += 7) {
                        if (pos > 64) {
                            aLine = reader.readLine();
                            if (aLine == null)
                                throw new IOException(format(RESOURCES.getString("EOFErrMsg"), reader.getLineNumber()));
                            pos = 7;
                        }

                        buffer = aLine.substring(pos, pos + 7).trim();
                        value = nf.parse(buffer).doubleValue();
                        depData[j][i] = value;
                    }
                }

                //  Determine the table name.
                switch (tbl) {
                    case 0:
                        tableName = "Cl";
                        break;
                    case 1:
                        tableName = "Cd";
                        break;
                    default:
                        tableName = "Cm";
                        break;
                }

                //  Create a new table object and add to database.
                Float64Table table = Float64Table.valueOf(tableName, indepNames, machBP, alphaBP, depData);
                db.add(table);

            }

        } catch (ParseException e) {
            throw new IOException(format(RESOURCES.getString("parseErrMsg"), reader.getLineNumber()), e);
        }

        return db;
    }

}
