/*
 * FTableReader  -- A class for reading table files of various formats.
 *
 * Copyright (C) 1999-2016 by Joseph A. Huwaldt
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *   
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.tables;

import java.io.IOException;
import java.io.InputStream;
import java.io.BufferedInputStream;
import java.io.OutputStream;
import java.util.ResourceBundle;

/**
 * This interface provides a standard interface for reading and writing
 * {@link Float64Table table} data from/to streams in various formats.
 * 
 * <p> Modified by: Joseph A. Huwaldt </p>
 * 
 * @author Joseph A. Huwaldt, Date: April 3, 2000
 * @version September 1, 2016
 */
public interface FTableReader extends Comparable<FTableReader> {

    /**
     * The resource bundle for this package.
     */
    static final ResourceBundle RESOURCES = Float64Table.RESOURCES;

    /**
     * Constant indicating that a reader can not read a specified file stream.
     */
    public static final int NO = 0;

    /**
     * Constant indicating that a reader can certainly read a specified file stream.
     */
    public static final int YES = 1;

    /**
     * Constant indicating that a reader might be able to read a specified file stream,
     * but can't determine for sure.
     */
    public static final int MAYBE = -1;

    /**
     * Returns the preferred file extension (not including the ".") for files of this
     * reader's type.
     *
     * @return The preferred file extension
     */
    public String getExtension();

    /**
     * Method that determines if this reader can read data from the specified input
     * stream.
     *
     * @param name  The name of the file where the input stream originated. This is used
     *              for examining file extensions to determine file type.
     * @param input An input stream containing the data to be read. Any methods that read
     *              from this stream must first set a mark and then reset back to that
     *              mark before the method returns (even if it returns with an exception).
     * @return FTableReader.NO if the file format is not recognized by this reader.
     *         FTableReader.YES if the file format is definitely recognized by this
     *         reader. FTableReader.MAYBE if the file format might be readable by this
     *         reader, but that can't easily be determined without actually reading the
     *         file.
     * @throws java.io.IOException
     */
    public int canReadData(String name, BufferedInputStream input) throws IOException;

    /**
     * Returns true if this class can write at least some data in the format supported by
     * this class. Returns false if it can not.
     *
     * @return true if this class can write at least some data in the format supported by
     *         this class.
     */
    public boolean canWriteData();

    /**
     * Interface for reading in data from an input stream and returning that data as a
     * list of {@link Float64Table} objects.
     *
     * @param input An input stream containing the data to be read.
     * @return A list of {@link Float64Table} objects that contain the data read in from
     *         the specified stream.
     * @throws IOException If there is a problem reading the specified stream.
     */
    public FTableList read(InputStream input) throws IOException;

    /**
     * Interface for writing out all the data stored in the specified list of
     * {@link Float64Table} objects to the specified output stream.
     *
     * @param output The output stream to which the data is to be written.
     * @param tables The list of {@link Float64Table} objects to be written out.
     * @throws IOException If there is a problem writing to the specified stream.
     */
    public void write(OutputStream output, FTableList tables) throws IOException;

}
