/*
 *   OTISTableReader -- A class that can read and write OTIS table (*.itd or *.dat) multi-table files.
 *
 *   Copyright (C) 2008-2024, by Joseph A. Huwaldt.
 *   All rights reserved.
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *  Or visit:  http://www.gnu.org/licenses/lgpl.html
 */
package jahuwaldt.js.tables;

import jahuwaldt.io.FileUtils;
import static jahuwaldt.js.tables.FTableReader.RESOURCES;
import jahuwaldt.js.util.TextTokenizer;
import java.io.*;
import java.text.DateFormat;
import java.text.MessageFormat;
import static java.text.MessageFormat.format;
import java.util.Date;
import java.util.Locale;
import java.util.NoSuchElementException;
import static java.util.Objects.isNull;
import static java.util.Objects.requireNonNull;
import java.util.logging.Level;
import java.util.logging.Logger;
import javolution.text.CharSet;
import javolution.text.Text;
import javolution.text.TypeFormat;
import javolution.util.FastMap;
import javolution.util.FastTable;
import org.jscience.mathematics.number.Float64;
import org.jscience.mathematics.number.Integer64;
import org.jscience.mathematics.vector.Float64Vector;

/**
 * <p>
 * A class that provides methods for reading or writing an OTIS table (*.itd or *.dat)
 * formatted multi-table file. These files are used by the OTIS trajectory simulation
 * system to store aerodynamic and propulsion data among other things. See the OTIS
 * documentation for more information.
 * </p>
 * <p>
 * When writing a table, the curve fit degree (0, 1, 3, or 5) can be set by placing an
 * integer in the table meta-data with the key <code>CurveFitDegree_INDEPNAME</code>;
 * where "INDEPNAME" is replaced with the name of one of the independent variables. A
 * chamfer can be specified by setting the "CurveFitDegree" to 6 and setting
 * <code>OTISChamferWidth_INDEPNAME</code> to a double value (it defaults to 0.05).
 * Finally, the table descriptor can be set by placing a character sequence (such as a
 * String) into the meta-data with the key "OTISDescriptor". When reading an OTIS table
 * file, those meta-data parameters will be set by the values read from the file.
 * </p>
 *
 * <p> Modified by: Joseph A. Huwaldt </p>
 *
 * @author Joseph A. Huwaldt, Date: January 10, 2008
 * @version January 15, 2024
 */
public class OTISTableReader implements FTableReader {

    //  The preferred file extension for files of this reader's type.
    private static final String EXTENSION = "itd";

    //  The character used to indicate that a line is commented out.
    private static final String COMMENT = "!";

    //  The delimiters used when parsing the file.
    private static final String DELIMITERS = "\t ,";

    //  The code that signifies the end of the table data.
    private static final Text END = Text.intern("END"); //  Do not write out an END code.

    //  Wrap output after this many columns of numbers.
    private static final int MAX_COLUMNS = 10;

    //  The max. character length for parameters.
    private static final int MAX_CHAR = 30;

    /**
     * The table meta-data key for retrieving or storing the table descriptor.
     */
    public static final String OTISDESCRIPTOR_KEY = "OTISDescriptor";

    /**
     * The table meta-data key for retrieving or storing the curve fit degree.
     */
    public static final String CURVEFITDEGREE_KEY = "CurveFitDegree";

    /**
     * The table meta-data key for retrieving or storing the chamfer width.
     */
    public static final String CHAMFERWIDTH_KEY = "OTISChamferWidth";

    /**
     * Returns a string representation of the object. This will return a brief description
     * of the format read by this reader.
     */
    @Override
    public String toString() {
        return RESOURCES.getString("otisDescription");
    }

    /**
     * Returns the preferred file extension (not including the ".") for files of this
     * reader's type.
     *
     * @return the preferred file extension of "itd".
     */
    @Override
    public String getExtension() {
        return EXTENSION;
    }

    /**
     * Method that determines if this reader can read data from the specified input
     * stream. Will return FTableReader.YES if the stream is recognized as an OTIS table
     * file.
     *
     * @param name  The name of the file.
     * @param input An input stream containing the data to be read. Any methods that read
     *              from this stream must first set a mark and then reset back to that
     *              mark before the method returns (even if it returns with an exception).
     * @return FTableReader.NO if the file is not recognized at all or FTableReader.MAYBE
     * if the filename extension is "*.itd" or "*.dat".
     * @throws java.io.IOException
     */
    @Override
    public int canReadData(String name, BufferedInputStream input) throws IOException {
        int response = NO;

        String ext = FileUtils.getExtension(requireNonNull(name, MessageFormat.format(RESOURCES.getString("paramNullErr"), "name")));
        if ("dat".equals(ext) || EXTENSION.equals(ext))
            response = MAYBE;

        return response;
    }

    /**
     * Returns true. This class can write data to an OTIS table file.
     *
     * @return true
     */
    @Override
    public boolean canWriteData() {
        return true;
    }

    /**
     * Compares this object with the specified object for order based on the
     * <code>toString().compareTo(o.toString())</code> method. Returns a negative integer,
     * zero, or a positive integer as this object is less than, equal to, or greater than
     * the specified object.
     */
    @Override
    public int compareTo(FTableReader o) {
        requireNonNull(o, MessageFormat.format(RESOURCES.getString("paramNullErr"), "o"));
        return this.toString().compareTo(o.toString());
    }

    /**
     * Method that reads in OTIS table formatted data from the specified input stream and
     * returns that data as an {@link FTableList} object.
     *
     * @param input An input stream containing the OTIS table formatted data.
     * @return An {@link FTableList} object that contains the data read in from the
     * specified stream.
     * @throws IOException If there is a problem reading the specified stream.
     */
    @Override
    public FTableList read(InputStream input) throws IOException {
        requireNonNull(input, MessageFormat.format(RESOURCES.getString("paramNullErr"), "input"));

        //  OTIS files are required to be in ASCII with U.S. style number formatting.
        //  Get the default locale and save it.  Then change the default locale to US.
        Locale defLocale = Locale.getDefault();
        Locale.setDefault(Locale.US);

        try {
            // Create a reader to translate the input stream.
            LineNumberReader reader = new LineNumberReader(new InputStreamReader(input));

            FTableList db = readStream(reader);

            return db;

        } finally {
            //  Restore the default locale.
            Locale.setDefault(defLocale);
        }
    }

    /**
     * Method for writing out all the data stored in the specified list of
     * {@link Float64Table} objects to the specified output stream in OTIS table format.
     *
     * @param output The output stream to which the data is to be written.
     * @param tables The list of Float64Table objects to be written out.
     * @throws IOException Always thrown as writing of this format is not yet supported.
     */
    @Override
    public void write(OutputStream output, FTableList tables) throws IOException {
        requireNonNull(output, MessageFormat.format(RESOURCES.getString("paramNullErr"), "output"));
        requireNonNull(tables, MessageFormat.format(RESOURCES.getString("paramNullErr"), "tables"));
        if (tables.isEmpty())
            return;

        //  OTIS files are required to be in ASCII with U.S. style number formatting.
        //  Get the default locale and save it.  Then change the default locale to US.
        Locale defLocale = Locale.getDefault();
        Locale.setDefault(Locale.US);

        try {
            //  Wrap the output stream in a PrintWriter.
            PrintWriter out = new PrintWriter(output);

            //  Write out any file level notes first.
            writeFileComments(out, tables);

            //  Loop over all the tables in the table list, writing out each one in turn.
            for (Float64Table table : tables) {
                writeTable(out, table);
            }

            //  Write "END" at the end.
            //out.println(END);
            
            //  Flush the output stream before leaving.
            out.flush();

        } finally {
            //  Restore the default locale.
            Locale.setDefault(defLocale);
        }
    }

    /**
     * A set of characters that are illegal to OTIS. The characters in this set include: '
     * ',/,\,(,),{,},\n,\r,\t,\0,\f,',?,*,<,>,|,",:,~,@,!,#,[,],=,+,;, and ','.
     */
    private static final CharSet ILLEGAL_CHARACTERS
            = CharSet.valueOf(' ', '/', '\\', '\n', '\r', '\t', '\0', '\f', '`', '?', '*', '<', '>', '|', '\"', ':', '~', '@', '!', '#',
                    '[', ']', '=', '+', ';', ',', '(', ')', '{', '}');

    /**
     * Sanitizes a table name to be compatible with OTIS. This method will replace any
     * illegal characters with "_". This also truncates the length of the name to 8
     * characters.
     *
     * @param name The table name to be sanitized.
     * @return A new table name that has been sanitized to be OTIS compatible.
     */
    public static Text sanitizeTableName(CharSequence name) {
        if (isNull(name))
            return null;
        Text txt = Text.valueOf(name);
        txt = txt.replace(ILLEGAL_CHARACTERS, "_");
        if (txt.length() > 8)
            txt = txt.subtext(0, 8);
        return txt;
    }

    /**
     * Sanitizes an independent variable name to be compatible with OTIS. This method will
     * replace any illegal characters with "_". This also truncates the length of the name
     * to 16 characters and converts the name to upper case.
     *
     * @param name an independent variable name to be sanitized.
     * @return A new independent name that has been sanitized to be OTIS compatible.
     */
    public static Text sanitizeIndepName(CharSequence name) {
        if (isNull(name))
            return null;
        Text txt = Text.valueOf(name);
        txt = txt.replace(ILLEGAL_CHARACTERS, "_");
        if (txt.length() > 8)
            txt = txt.subtext(0, 8);
        return txt.toUpperCase();
    }

    /**
     * Read in an OTIS table formatted multi-table stream and create a new table database
     * with the new files.
     *
     * @param in A reader that reads in the characters from the input stream.
     * @return A table database containing the tables read in from the stream.
     */
    private static FTableList readStream(LineNumberReader in) throws IOException {
        
        // Create an empty table database.
        FTableList db = FTableList.newInstance();

        //  Read in the 1st line to see if there is any data.
        in.mark(10240);
        Text aLine = readLineNoComments(in);
        if (aLine == null)
            throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));

        while (aLine != null && !aLine.trim().equals(END)) {
            in.reset();

            //  Read a single table from the file.
            Float64Table theTable = readSingleTable(in);

            //  Add this table to the table database.
            if (theTable != null)
                db.add(theTable);

            //  Is there more data?
            in.mark(10240);
            aLine = readLineNoComments(in);

        }

        return db;
    }

    /**
     * Read a single table from an OTIS formatted table file.
     *
     * @param in A file reader that reads in the characters from the file.
     * @return A table containing the data read in from the disk file.
     */
    private static Float64Table readSingleTable(LineNumberReader in) throws IOException {
        Float64Table theTable = null;

        try {
            //  Read in the table name.
            Text aLine = readFixedWidthLineNoComments(in, MAX_CHAR);
            if (aLine == null)
                throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));
            Text tblName = aLine.trim();

            //  Read the table descriptor.
            aLine = readFixedWidthLineNoComments(in, MAX_CHAR);
            if (aLine == null)
                throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));
            Text descriptor = aLine.trim();

            //  Read in the scale factor.
            aLine = readFixedWidthLineNoComments(in, MAX_CHAR);
            if (aLine == null)
                throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));
            double scaleF = TypeFormat.parseDouble(aLine.trim());

            //  Read in the number of terms.
            aLine = readFixedWidthLineNoComments(in, MAX_CHAR);
            if (aLine == null)
                throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));
            int numTerms = TypeFormat.parseInt(aLine.trim());
            if (numTerms == 0) {
                Logger.getLogger(OTISTableReader.class.getName()).log(Level.WARNING,
                        format(RESOURCES.getString("otisNoConstantsMsg"), tblName));
                return null;
            }
            if (numTerms != 1)
                throw new IOException(format(RESOURCES.getString("otisNoMultTermTablesErr"), tblName));

            //  Skip direct variable multiplier.
            aLine = readFixedWidthLineNoComments(in, MAX_CHAR);
            if (aLine == null)
                throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));
            String[] parts = aLine.trim().toString().split(" ");
            parts[0] = parts[0].trim().toUpperCase();
            if (parts[0].equals("NDVM")) {
                int ndvm = TypeFormat.parseInt(parts[1].trim());
                if (ndvm != 0)
                    throw new IOException(format(RESOURCES.getString("otisNoDVMTablesErr"), tblName));
            }

            //  Skip in the number of term 1 coefficients.
            if (!parts[0].equals("NCOEF")) {
                aLine = readLineNoComments(in);
                if (aLine == null)
                    throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));
            }

            //  Read in the number of independent variables in the table.
            aLine = readFixedWidthLineNoComments(in, MAX_CHAR);
            if (aLine == null)
                throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));
            int numIndep = TypeFormat.parseInt(aLine.trim());
            if (numIndep < 1)
                throw new IOException(format(RESOURCES.getString("otisNoIndepVarErr"), tblName));

            //  Loop over all the independent variables.
            FastTable<Text> indepNames = FastTable.newInstance();
            FastTable<Integer64> elemPerDim = FastTable.newInstance();
            FastTable<Float64Vector> breakpoints = FastTable.newInstance();
            FastMap<String, Object> tblMetaData = FastMap.newInstance();
            TextTokenizer tokenizer = TextTokenizer.newInstance();
            tokenizer.setDelimiters(DELIMITERS);
            for (int i = 0; i < numIndep; ++i) {

                //  Read in the name of the independent variable.
                aLine = readFixedWidthLineNoComments(in, MAX_CHAR);
                if (aLine == null)
                    throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));
                Text indepName = aLine.trim();
                indepNames.add(indepName);

                //  Read in the number of breakpoints and curve fit type for this independent variable.
                aLine = readFixedWidthLineNoComments(in, MAX_CHAR);
                if (aLine == null)
                    throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));
                tokenizer.setText(aLine);
                Text token = tokenizer.next();
                double nvl = TypeFormat.parseDouble(token);
                int numBP = (int) nvl;                       //  Number of breakpoints for this independent variable.
                int cfDeg = (int) Math.round((nvl - numBP) * 10);        //  The curve fit degree.
                if (cfDeg == 0)
                    cfDeg = 5;
                if (numBP < 0) {
                    numBP *= -1;
                    cfDeg = 0;
                }
                double Wc = 0.05;
                if (cfDeg == 6 && tokenizer.hasNext()) {
                    //  Read in the chamfer width.
                    token = tokenizer.next();
                    Wc = TypeFormat.parseDouble(token);
                }

                //  Store off the data.
                elemPerDim.add(Integer64.valueOf(numBP));
                tblMetaData.put(CURVEFITDEGREE_KEY + "_" + indepName, cfDeg);
                if (Wc != 0.05)
                    tblMetaData.put(CHAMFERWIDTH_KEY + "_" + indepName, Wc);
                FastTable<Float64> dimBreakpoints = FastTable.newInstance();

                //  Loop over and read in all the breakpoints.
                int count = 0;
                while (count < numBP) {
                    aLine = readLineNoComments(in);
                    if (aLine == null)
                        throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));

                    tokenizer.setText(aLine);
                    for (Text atoken : tokenizer) {
                        atoken = atoken.trim();
                        Float64 value = Float64.valueOf(atoken);
                        dimBreakpoints.add(value);
                        ++count;
                    }

                }

                breakpoints.add(Float64Vector.valueOf(dimBreakpoints));
                FastTable.recycle(dimBreakpoints);
            }   //  Next i
            TextTokenizer.recycle(tokenizer);

            //  Create an n-dimensional array to hold the dependent data.
            Float64NDimArray depData = Float64NDimArray.newInstance(elemPerDim);
            FastTable.recycle(elemPerDim);

            //  Recursively read in the dependent data.
            FastTable<Integer64> indep = FastTable.newInstance();
            for (int i = 0; i < numIndep; ++i)
                indep.add(Integer64.ZERO);
            rReadData(numIndep - 1, in, indep, depData, breakpoints, scaleF);

            // Create a new floatTable object using this information.
            theTable = Float64Table.valueOf(tblName, indepNames, breakpoints, depData);

            //  Add some notes to the table.
            tblMetaData.put(OTISDESCRIPTOR_KEY, descriptor);
            theTable.putAllMetaData(tblMetaData);

            //  Recycle the tables.
            FastTable.recycle(indepNames);
            FastTable.recycle(breakpoints);
            FastTable.recycle(indep);
            Float64NDimArray.recycle(depData);
            FastMap.recycle(tblMetaData);

        } catch (NoSuchElementException | NumberFormatException e) {
            throw new IOException(format(RESOURCES.getString("parseErrMsg"), in.getLineNumber()), e);
        }

        return theTable;
    }

    /**
     * Recursively read in arrays of dependent data from an OTIS table formatted stream
     * into the parameter depData.
     *
     * @param dim         The dimension (independent) we are currently looping through.
     *                    Should initially be set to the number of independent variables -
     *                    1.
     * @param in          A file reader that reads in the characters from the file.
     * @param indep       A list of indexes into the data table (independents).
     * @param depData     The dependent data array we are reading into.
     * @param breakpoints The array of breakpoint arrays.
     */
    private static void rReadData(int dim, LineNumberReader in, FastTable<Integer64> indep, Float64NDimArray depData,
            FastTable<Float64Vector> breakpoints, double scaleF)
            throws IOException, NumberFormatException, NoSuchElementException {
        if (dim == 0) {
            // Reached the lowest dimension, indep[] is now ready.
            // Read in an array.
            //for (int i=1; i < indep.size(); ++i)
            //  System.out.println("indep(" + i + ") = " + indep.get(i));
            readDataArray(in, indep, depData, breakpoints.get(dim).getDimension(), scaleF);

        } else {

            // Loop over all the breakpoints in this dimension.
            for (int i = 0; i < breakpoints.get(dim).getDimension(); ++i) {

                // Fill in the required value for indep[] at this dimension.
                indep.set(dim, Integer64.valueOf(i));

                // Go on to the next lower dimension.
                rReadData(dim - 1, in, indep, depData, breakpoints, scaleF);
            }
        }
    }

    /**
     * Read a single array of dependent values from an OTIS table file stream. The data
     * read in is stored in the parameter depData.
     *
     * @param in      A file reader that reads in the characters from the file.
     * @param indep   A list of indexes into the data table (independents).
     * @param depData The array of dependent values we are reading into.
     * @param length  The length of most rapidly varying breakpoint array (
     *                breakpoints.get(numDims-1).getDimension() ).
     * @param scaleF  A scale factor used to pre-multiply all the data in the table.
     */
    private static void readDataArray(LineNumberReader in, FastTable<Integer64> indep,
            Float64NDimArray depData, int length, double scaleF)
            throws IOException, NumberFormatException, NoSuchElementException {

        TextTokenizer tokenizer = TextTokenizer.newInstance();
        tokenizer.setDelimiters(DELIMITERS);

        try {
            // Loop over all the breakpoints in the last independent (the array) and read in the array.
            int count = 0;
            while (count < length) {
                Text aLine = readLineNoComments(in);
                if (aLine == null)
                    throw new IOException(format(RESOURCES.getString("EOFErrMsg"), in.getLineNumber()));

                tokenizer.setText(aLine);
                for (Text token : tokenizer) {
                    token = token.trim();
                    double value = TypeFormat.parseDouble(token);
                    value *= scaleF;
                    indep.set(0, Integer64.valueOf(count));     //  indep[0] = count;
                    depData.set(indep, value);
                    ++count;
                }
            }

        } catch (ArrayIndexOutOfBoundsException e) {
            throw new IOException(format(RESOURCES.getString("otisUnexpNumElementsErr"), in.getLineNumber()), e);

        } finally {
            TextTokenizer.recycle(tokenizer);
        }
    }

    /**
     * Reads a line of text from the supplied file reader stripping out any lines that are
     * commented out.
     */
    private static Text readLineNoComments(LineNumberReader reader) throws IOException {

        String aLine = reader.readLine();
        while (aLine != null && (aLine.startsWith(COMMENT) || aLine.startsWith("*") || aLine.trim().equals("")))
            aLine = reader.readLine();

        return (aLine != null ? Text.valueOf(aLine) : null);
    }

    /**
     * Reads in a line of text from the supplied file reader stripping out any lines that
     * are commented out and returning on the "maxCols" number of characters from the
     * line.
     *
     * @param reader A reference to a LineNumberReader for the file.
     * @param maxWidth The maximum width of the Text to return.
     * @return
     * @throws IOException
     */
    private static Text readFixedWidthLineNoComments(LineNumberReader reader, int maxWidth) throws IOException {
        Text aLine = readLineNoComments(reader);
        if (aLine != null) {
            //  Keep only the 1st "maxCol" characters.
            aLine = aLine.subtext(0, (aLine.length() > maxWidth ? maxWidth : aLine.length()));
        }
        return aLine;
    }

    /**
     * Write out any database notes as file level comments to the OTIS table file.
     *
     * @param output A writer used to output the data.
     * @param tables The table database we are writing out.
     */
    private static void writeFileComments(PrintWriter output, FTableList tables)
            throws IOException {
        int numNotes = tables.numberOfNotes();
        if (numNotes == 0) {

            // Write out some generic table comments into the header lines.
            output.print(COMMENT);
            output.println("******************************************");
            output.print(COMMENT);
            output.print(" ");
            tables.toText().println(output);
            output.println(COMMENT);
            output.print(COMMENT);
            output.print(" Date: ");
            Date theDate = new Date();
            output.print(DateFormat.getDateInstance(DateFormat.SHORT).format(theDate));
            output.print(", ");
            output.println(DateFormat.getTimeInstance().format(theDate));
            output.print(COMMENT);
            output.println("******************************************");

        } else {
            // Write out the table notes as the header of the file.
            output.print(COMMENT);
            output.println("******************************************");

            for (int i = 0; i < numNotes; ++i) {
                Text note = tables.getNote(i);
                if (!note.startsWith(COMMENT)) {
                    output.print(COMMENT);
                    output.print(" ");
                }
                output.println(note.toString());
            }

            output.print(COMMENT);
            output.println("******************************************");
        }
    }

    /**
     * Write out a single table to the output print stream.
     */
    private void writeTable(PrintWriter out, Float64Table table) throws IOException {
        if (table == null) {
            System.out.println("Null table!");
            return;
        }

        //  Write out any table notes first.
        int numNotes = table.numberOfNotes();
        if (numNotes > 0) {
            out.print(COMMENT);
            out.println("******************************************");
            for (Text note : table.getAllNotes()) {
                if (!note.startsWith(COMMENT)) {
                    out.print(COMMENT);
                    out.print(" ");
                }
                out.println(note.toString());
            }
            out.print(COMMENT);
            out.println("******************************************");
        }

        //  Write out the table name.
        Text tblName = sanitizeTableName(table.getName());
        out.println(tblName);

        //  Write out the table descriptor (if there is one), otherwise just the table name again.
        Object descriptor = table.getMetaData(OTISDESCRIPTOR_KEY);
        if (descriptor != null) {
            out.println(sanitizeTableName(Text.valueOf(descriptor)));
        } else
            out.println(tblName);

        //  Write out the scale factor (always 1).
        out.println("1.");

        //  Write out the number of terms (always 1).
        out.println("1");

        //  Write out the number of tabular coefficients.
        out.println("NCOEF 1");

        //  Write out the number of dimensions.
        int numIndep = table.getDimensions();
        if (numIndep > 5)
            throw new IOException(format(RESOURCES.getString("otisToManyDimsErr")));
        out.println(numIndep);

        //  Loop over the dimensions and output the breakpoints.
        FastTable<Integer64> elemPerDim = FastTable.newInstance();
        FastTable<Float64Vector> breakpoints = FastTable.newInstance();
        for (int i = 0; i < numIndep; ++i) {

            //  Write out the name of this independent variable.
            Text indepName = sanitizeIndepName(table.getIndepName(i));
            out.println(indepName);

            //  Get the number of breakpoints in this dimension.
            int numBP = table.getNumBreakpoints(i);
            elemPerDim.add(Integer64.valueOf(numBP));

            //  Determine the curve fit degree.
            String key = CURVEFITDEGREE_KEY + "_" + indepName;
            Integer cfDeg = (Integer)table.getMetaData(key);
            if (cfDeg == null)
                cfDeg = 1;  // Default to linear.

            //  Output the number of breakpoints (negative indicates the degree is zero or floor).
            if (cfDeg == 0)
                out.print("-");
            out.print(numBP);

            //  Output the curve fit degree if it is anything other than 5 (which is the OTIS default).
            if (cfDeg != 6 && cfDeg != 0) {
                out.print(".");
                out.print(cfDeg);
            }
            if (cfDeg == 6) {
                //  Look for a chamfer width.
                key = CHAMFERWIDTH_KEY + "_" + indepName;
                Double Wc = (Double)table.getMetaData(key);
                if (Wc == null)
                    Wc = 0.05;
                //  Output the chamfer width.
                if (Wc != 0.05) {
                    out.print(" ");
                    out.print(Wc);
                }
            }
            out.println();

            //  Write out the breakpoints for this dimension.
            Float64Vector dimBreakpoints = table.getBreakpoints(i);
            breakpoints.add(dimBreakpoints);
            int icol = 1;
            for (int j = 0; j < numBP; ++j) {
                double bp = dimBreakpoints.getValue(j);
                out.printf("%10g ", bp);
                if (icol == MAX_COLUMNS) {
                    out.println();
                    icol = 0;
                }
                ++icol;
            }
            if (icol != 1)
                out.println();      //  Output a line feed if necessary.
        }

        //  Recursively write out the dependent data.
        FastTable<Integer64> indep = FastTable.newInstance();
        for (int i = 0; i < numIndep; ++i)
            indep.add(Integer64.ZERO);
        rWriteData(numIndep - 1, out, indep, table.getAll(), breakpoints);

        //  Put a blank line in.
        out.println(COMMENT);
    }

    /**
     * Recursively write out arrays of dependent data to an OTIS table formatted stream.
     *
     * @param dim         The dimension (independent) we are currently looping through.
     *                    Should initially be set to the number of independent variables -
     *                    1.
     * @param out         A PrintWriter that writes the characters to the file.
     * @param indep       A list of indexes into the data table (independents).
     * @param depData     The dependent data array we are writing from.
     * @param breakpoints The array of breakpoint arrays.
     */
    private static void rWriteData(int dim, PrintWriter out, FastTable<Integer64> indep, Float64NDimArray depData,
            FastTable<Float64Vector> breakpoints) throws IOException {
        if (dim == 0) {
            // Reached the lowest dimension, indep[] is now ready.
            // Read in an array.
            writeDataArray(out, indep, depData, breakpoints.get(dim).getDimension());

        } else {

            // Loop over all the breakpoints in this dimension.
            for (int i = 0; i < breakpoints.get(dim).getDimension(); ++i) {

                // Fill in the required value for indep[] at this dimension.
                indep.set(dim, Integer64.valueOf(i));

                // Go on to the next lower dimension.
                rWriteData(dim - 1, out, indep, depData, breakpoints);
            }
        }
    }

    /**
     * Write a single array of dependent values to an OTIS table file stream.
     *
     * @param out     A PrintWriter that writes the characters to the file.
     * @param indep   A list of indexes into the data table (independents).
     * @param depData The array of dependent values we are writing out.
     * @param length  The length of most rapidly varying breakpoint array (
     *                breakpoints.get(numDims-1).getDimension() ).
     */
    private static void writeDataArray(PrintWriter out, FastTable<Integer64> indep,
            Float64NDimArray depData, int length) throws IOException {

        // Loop over all the breakpoints in the last independent (the array) and write the array.
        int icol = 1;
        for (int i = 0; i < length; ++i) {
            indep.set(0, Integer64.valueOf(i));     //  indep[0] = i;
            double value = depData.getValue(indep);
            out.printf("%10g ", value);
            if (icol == MAX_COLUMNS) {
                out.println();
                icol = 0;
            }
            ++icol;
        }
        if (icol != 1)
            out.println();      //  Output a line feed if necessary.
    }

}
