/**
*   Please feel free to use any fragment of the code in this file that you need
*   in your own work. As far as I am concerned, it's in the public domain. No
*   permission is necessary or required. Credit is always appreciated if you
*   use a large chunk or base a significant product on one of my examples,
*   but that's not required either.
*
*   This code is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*
*      --- Joseph A. Huwaldt
**/
package jahuwaldt.swing;

import java.awt.Image;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;


/**
*  A <code>Transferable</code> which implements the capability required
*  to transfer an <code>Image</code>. This <code>Transferable</code> properly supports
*  <code>DataFlavor.imageFlavor</code> and all equivalent flavors.
*  No other DataFlavors are supported.
*
*  <p>  Modified by:  Joseph A. Huwaldt   </p>
*
*  @author   Joseph A. Huwaldt   Date:  March 5, 2009
*  @version  March 5, 2009
**/
public class ImageSelection implements Transferable, ClipboardOwner {

	//	The image being transferred.
    private Image image;

	/**
	*  Creates a <code>Transferable</code> capable of transferring the specified <code>Image</code>.
	**/
    public ImageSelection(Image image) {
        this.image = image;
    }

    /**
	*  Returns an array of flavors in which this <code>Transferable</code>
	*  can provide the data. <code>DataFlavor.imageFlavor</code> is properly supported.
	*
	*  @return an array of length one, whose only element is <code>DataFlavor.imageFlavor</code>
	**/
    public synchronized DataFlavor[] getTransferDataFlavors() {
        return new DataFlavor[]{DataFlavor.imageFlavor};
    }

    /**
	*  Returns whether the requested flavor is supported by this <code>Transferable</code>.
	*
	*  @param flavor  the requested flavor for the data
	*  @return true if <code>flavor</code> is equal to <code>DataFlavor.imageFlavor</code>;
	*          false if <code>flavor</code> is any other flavor.
	**/
    public boolean isDataFlavorSupported(DataFlavor flavor) {
        return DataFlavor.imageFlavor.equals(flavor);
    }

    /**
	*  Returns the <code>Transferable</code>'s data in the requested <code>DataFlavor</code> if possible.
	*  If the desired flavor is <code>DataFlavor.imageFlavor</code>, or an equivalent flavor, the 
	*  <code>Image</code> representing the selection is returned.
	*
	*  @param flavor  the requested flavor for the data
	*  @return the data in the requested flavor, as outlined above
	*  @throws UnsupportedFlavorException if the requested data flavor is not equivalent to DataFlavor.imageFlavor
	*  @throws IOException if the data is no longer available in the requested flavor.
	*  @throws NullPointerException if flavor is <code>null</code>
	**/
    public synchronized Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException, IOException {
        if (DataFlavor.imageFlavor.equals(flavor))
			return image;
		else
            throw new UnsupportedFlavorException(flavor);
    }
	
	/**
	* Notifies this object that it is no longer the clipboard owner.
	* This method will be called when another application or another object within
	* this application asserts ownership of the clipboard.
	**/
	public void lostOwnership(Clipboard clipboard, Transferable contents) { }
	
}